%{ open Ast

let block_id = ref 1

let gen_block_id = function
	0 -> let x = block_id.contents in
		block_id := x + 1; x
	| x -> raise(Failure("internal error")) (* fixme - blocks are not numbering right*)
	
%}

%token LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET EOF
%token PLUS MINUS TIMES DIVIDE MOD
%token OR AND
%token LESS GREATER LESSEQUAL GREATEREQUAL EQ NEQ NOT
%token TILDE
%token ASSIGN

%token IF ELSE WHILE RET

%token RET

%token SEMI COMMA

%token STR NUM VOID MAP

%token <string> STR_LITERAL
%token <int> NUM_LITERAL

%token <string> ID

%nonassoc NOELSE

%left COMMA

%right ASSIGN

%left OR 
%left AND 

%left EQ NEQ
%left LESS LESSEQUAL GREATER GREATEREQUAL 

%left PLUS MINUS
%left TIMES DIVIDE MOD
%left TILDE

%left NEG NOT
%left LBRACKET

%start program
%type <Ast.program> program

%%

program:
	func_list { { globals = []; functions = List.rev $1; block_count = gen_block_id 0} }

func_list:
	{ [] }
	| func_list func { $2 :: $1 }

func:
	var_type ID LPAREN formals_list_opt RPAREN block
		{ { name = $2; ret_type = $1; body = $6; formals = $4} }

block:
	LBRACE var_list /*func_list*/ stmt_list RBRACE
		{ {locals = List.rev $2; subfunctions = []; statements = $3; block_id = gen_block_id 0} }

stmt_list:
	{ [] }
	| stmt stmt_list { $1 :: $2 }

stmt:
	block { CodeBlock($1) }
	| expr SEMI { Expression($1) }
	| WHILE LPAREN expr RPAREN block { Loop($3, $5) }
	| IF LPAREN expr RPAREN block ELSE block { Conditional($3, $5, $7) }
	| IF LPAREN expr RPAREN block %prec NOELSE { Conditional($3, $5, {locals = []; subfunctions = []; statements = []; block_id = -1}) }
	| RET expr SEMI { Return($2) }
	| RET SEMI { Return(NoExpr) }

expr:
	/* binary arithmetic operators */
	expr PLUS expr { Binop($1, Plus, $3) }
	| expr MINUS expr { Binop($1, Minus, $3) }
	| expr TIMES expr { Binop($1, Times, $3) }
	| expr DIVIDE expr { Binop($1, Divide, $3) }
	| expr MOD expr { Binop($1, Mod, $3) }

	/* comparison */
	| expr LESS expr { Binop($1, Less, $3) }
	| expr GREATER expr { Binop($1, Greater, $3) }
	| expr LESSEQUAL expr { Binop($1, Leq, $3) }
	| expr GREATEREQUAL expr { Binop($1, Geq, $3) }

	| expr TILDE expr TILDE expr { Replace($1, $3, $5) }

	/* equality */
	| expr EQ expr { Binop($1, Eq, $3) }
	| expr NEQ expr { Binop($1, Neq, $3) }
	
	/* logical or/and */
	| expr OR expr { Binop($1, Or, $3) }
	| expr AND expr { Binop($1, And, $3) }

	/* assignment */	
	| lvalue ASSIGN expr { Assign($1, $3) }
	
	/* arithmetic and logical negation */
	| MINUS expr %prec NEG { Unop($2, Neg) }
	| NOT expr { Unop($2, Not) }
	
	/* grouping */
	| LPAREN expr RPAREN { $2 }

	/* function call */
	| ID LPAREN actuals_list_opt RPAREN { FuncCall($1, $3) }
	
	/* variable */
	| lvalue { Rvalue($1) }

	/* literals */
	| STR_LITERAL { StrLiteral($1) }
	| NUM_LITERAL { NumLiteral($1) }

formals_list_opt:
	formals_list { $1 }
	| VOID { [] }
	
formals_list:
	var { [$1] }
	| var COMMA formals_list { $1 :: $3 }

actuals_list_opt:
	{ [] }
	| actuals_list { $1 }

actuals_list:
	expr { [$1] }
	| expr COMMA actuals_list { $1 :: $3 }

var_list:
	{ [] }
	| var_list var SEMI { $2 :: $1 }

var:
	var_type ID { ($2, $1) }

var_type:
	STR { Simple(Str) }
	| NUM { Simple(Num) }
	| MAP LESS STR COMMA STR GREATER { Map(Str, Str) }
	| MAP LESS STR COMMA NUM GREATER { Map(Str, Num) }
	| MAP LESS NUM COMMA STR GREATER { Map(Num, Str) }
	| MAP LESS NUM COMMA NUM GREATER { Map(Num, Num) }

lvalue:
	ID { ($1, NoExpr) }
	| ID LBRACKET expr RBRACKET { ($1, $3) }
